@@ -1,7 +1,7 @@
 package App::shcompgen;
 
 our $DATE = '2016-10-25'; # DATE
-our $VERSION = '0.22'; # VERSION
+our $VERSION = '0.23'; # VERSION
 
 use 5.010001;
 use strict;
@@ -41,7 +41,7 @@ my $_complete_prog = sub {
     }
 };
 
-our @supported_shells = qw(bash fish zsh); # XXX tcsh
+our @supported_shells = qw(bash fish zsh tcsh);
 our %shell_arg = (
     shell => {
         summary => 'Override guessing and select shell manually',
@@ -94,15 +94,25 @@ _
         schema  => ['array*', of => 'str*'],
     },
 
-    #tcsh_global_dir => {
-    #    summary => 'Directory to put completions scripts',
-    #    schema  => ['array*', of => 'str*'],
-    #    default => '/etc/fish/completions',
-    #},
-    #tcsh_per_user_dir => {
-    #    summary => 'Directory to put completions scripts',
-    #    schema  => ['array*', of => 'str*'],
-    #},
+    tcsh_global_dir => {
+        summary => 'Directory to put completions scripts',
+        description => <<'_',
+
+Currently actually hardcoded to `/etc/tcsh/completions`.
+
+_
+        schema  => ['array*', of => 'str*'],
+        default => ['/etc/tcsh/completions'],
+    },
+    tcsh_per_user_dir => {
+        summary => 'Directory to put completions scripts',
+        description => <<'_',
+
+Currently actually hardcoded to `~/.config/tcsh/completions`.
+
+_
+        schema  => ['array*', of => 'str*'],
+    },
 
     zsh_global_dir => {
         summary => 'Directory to put completions scripts',
@@ -137,6 +147,7 @@ sub _set_args_defaults {
         my $sh = Shell::Guess->running_shell;
         my $n = $sh->{name};
         $n = "zsh" if $n eq 'z';
+        $n = "tcsh" if $n eq 'c';
         $n = "bash" if $n eq 'bourne'; # under make
         $args->{shell} = $n;
     }
@@ -152,13 +163,46 @@ sub _set_args_defaults {
     $args->{fish_global_dir}   //= ['/usr/share/fish/completions',
                                     '/etc/fish/completions'];
     $args->{fish_per_user_dir} //= ["$ENV{HOME}/.config/fish/completions"];
-    #$args->{tcsh_global_dir}   //= [];
-    #$args->{tcsh_per_user_dir} //= ["$ENV{HOME}/.config/tcsh/completions"];
+    $args->{tcsh_global_dir}   //= ['/etc/tcsh/completions'];
+    $args->{tcsh_per_user_dir} //= ["$ENV{HOME}/.config/tcsh/completions"];
     $args->{zsh_global_dir}    //= ['/usr/local/share/zsh/site-functions'];
     $args->{zsh_per_user_dir}  //= ["$ENV{HOME}/.config/zsh/completions"];
     [200];
 }
 
+sub _tcsh_init_script_path {
+    my %args = @_;
+    if ($args{global}) {
+        return "/etc/shcompgen.tcshrc";
+    } else {
+        return "$ENV{HOME}/.config/shcompgen.tcshrc";
+    }
+}
+
+sub _gen_tcsh_init_script {
+    my %args = @_;
+    my $dirs = $args{global} ?
+        $args{tcsh_global_dir} : $args{tcsh_per_user_dir};
+    my @defs;
+    for my $dir (@$dirs) {
+        next unless -d $dir;
+        for my $file (<$dir/*>) {
+            open my $fh, "<", $file or do {
+                warn "Can't open '$file': $!, skipped\n";
+                next;
+            };
+            my $line = <$fh>;
+            $line .= "\n" unless $line =~ /\n\z/;
+            push @defs, $line;
+        }
+    }
+    join(
+        "",
+        "# Generated by shcompgen on ", scalar(localtime), "\n",
+        @defs,
+    );
+}
+
 sub _gen_completion_script {
     require String::ShellQuote;
 
@@ -221,6 +265,16 @@ _|.$prog.q| "$@"
 |;
         }
 
+    } elsif ($shell eq 'tcsh') {
+
+        if (defined $args) {
+            $header_at_bottom++;
+            $script = "complete $qprog 'p/*/`$qcomp $args`/'\n";
+        } else {
+            $header_at_bottom++;
+            $script = "complete $qprog 'p/*/`$qcomp`/'\n";
+        }
+
     } elsif ($shell eq 'fish') {
 
         require File::Which;
@@ -505,6 +559,8 @@ sub _generate_or_remove {
     # to avoid writing a file and then removing the file again in the same run
     my %written_files;
 
+    my %removed_files;
+
     my $envres = envresmulti();
   PROG:
     for my $prog0 (@{ $args{prog} }) {
@@ -602,6 +658,7 @@ sub _generate_or_remove {
             $log->infof("Unlinking %s ...", $comppath);
             if (unlink $comppath) {
                 $envres->add_result(200, "OK", {item_id=>$prog0});
+                $removed_files{$comppath}++;
             } else {
                 $envres->add_result(500, "Can't unlink '$comppath': $!",
                                     {item_id=>$prog0});
@@ -610,6 +667,15 @@ sub _generate_or_remove {
 
     } # for prog0
 
+    if (keys(%written_files) || keys(%removed_files)) {
+        if ($args{shell} eq 'tcsh') {
+            my $init_script_path = _tcsh_init_script_path(%args);
+            my $init_script = _gen_tcsh_init_script(%args);
+            $log->debugf("Re-writing init script %s ...", $init_script_path);
+            write_text($init_script_path, $init_script);
+        }
+    }
+
     $envres->as_struct;
 }
 
@@ -763,6 +829,12 @@ _
             "\n\n" . " . $init_script_path\n\n";
     } elsif ($shell eq 'fish') {
         # nothing to do, ready by default
+    } elsif ($shell eq 'tcsh') {
+        $init_location = $global ? "/etc/csh.cshrc" : "~/.tcshrc";
+        $init_script = _gen_tcsh_init_script(%args);
+        $init_script_path = _tcsh_init_script_path(%args);
+        $instruction .= "Please put this into your $init_location:".
+            "\n\n" . " source $init_script_path\n\n";
     } else {
         return [412, "Shell '$shell' not yet supported"];
     }
@@ -952,7 +1024,7 @@ App::shcompgen - Generate shell completion scripts
 
 =head1 VERSION
 
-This document describes version 0.22 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-10-25.
+This document describes version 0.23 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-10-25.
 
 =head1 FUNCTIONS
 
@@ -1054,6 +1126,18 @@ Override guessing and select shell manually.
 
 Output completion script to STDOUT.
 
+=item * B<tcsh_global_dir> => I<array[str]> (default: ["/etc/tcsh/completions"])
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C</etc/tcsh/completions>.
+
+=item * B<tcsh_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C<~/.config/tcsh/completions>.
+
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
@@ -1141,6 +1225,18 @@ that and manually select which.
 
 Override guessing and select shell manually.
 
+=item * B<tcsh_global_dir> => I<array[str]> (default: ["/etc/tcsh/completions"])
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C</etc/tcsh/completions>.
+
+=item * B<tcsh_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C<~/.config/tcsh/completions>.
+
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
@@ -1207,6 +1303,18 @@ that and manually select which.
 
 Override guessing and select shell manually.
 
+=item * B<tcsh_global_dir> => I<array[str]> (default: ["/etc/tcsh/completions"])
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C</etc/tcsh/completions>.
+
+=item * B<tcsh_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C<~/.config/tcsh/completions>.
+
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
@@ -1278,6 +1386,18 @@ searched from PATH.
 
 Override guessing and select shell manually.
 
+=item * B<tcsh_global_dir> => I<array[str]> (default: ["/etc/tcsh/completions"])
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C</etc/tcsh/completions>.
+
+=item * B<tcsh_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
+Currently actually hardcoded to C<~/.config/tcsh/completions>.
+
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
